from polars import DataFrame
from sqlalchemy import text
from sqlalchemy.engine import Engine
from sqlalchemy.ext.asyncio import AsyncEngine


async def async_query_with_result(query: str, engine: AsyncEngine) -> DataFrame:
    async with engine.connect() as conn:
        result = await conn.execute(text(query))
        data = result.fetchall()

        return DataFrame(data=data)


def sync_query_with_result(query: str, engine: Engine) -> DataFrame:
    with engine.connect() as conn:
        result = conn.execute(text(query))
        data = result.fetchall()

        return DataFrame(data=data)



# mypy: disable_error_code="type-arg"

from typing import Any

from bson.objectid import ObjectId
from pymongo import MongoClient


def insert_into_collection(
    documents: list[dict[str, Any]],
    client: MongoClient,
    dbname: str,
    collection_name: str,
) -> list[ObjectId]:
    db = client[dbname]
    collection = db[collection_name]
    result = collection.insert_many(documents=documents)

    return result.inserted_ids



from datetime import datetime
from typing import Any, Sequence

import pandas as pd
import polars as pl
from pydantic import BaseModel, TypeAdapter, create_model


class Base(BaseModel):
    pass


async def camel_case(name: str) -> str:
    delimiters = []
    for character in name:
        if not character.isalnum():
            delimiters.append(character)

    for delimiter in delimiters:
        name = " ".join(name.split(delimiter))

    tokens = name.split()

    return "".join(token.lower().capitalize() for token in tokens)


async def deserialize_pydantic_objects(models: Sequence[BaseModel]) -> pl.DataFrame:
    data = [model.model_dump() for model in models]

    return pl.DataFrame(data=data)


async def flatten(data: pd.DataFrame | pl.DataFrame, depth: int) -> pl.DataFrame:
    if isinstance(data, pd.DataFrame):
        data = pl.from_pandas(data=data)

    return pl.json_normalize(data=data.to_dicts(), max_level=depth)


async def _serialize_data(model_name: str, data: list[dict[Any, Any]], fields: dict[str, Any]) -> list[BaseModel]:
    Model = create_model(model_name, **fields)

    ta = TypeAdapter(list[Model])  # type: ignore[valid-type]

    return ta.validate_python(data)


async def _serialize_table_polars(table_name: str, data: pl.DataFrame) -> list[BaseModel]:
    fields: dict[str, Any] = {}
    field_type: Any

    for column, dtype in data.schema.items():
        non_nulls = data.select(pl.col(column)).drop_nulls()
        cell = non_nulls.row(0)[0]

        if dtype.is_integer():
            field_type = int
        elif dtype.is_float():
            field_type = float
        elif dtype == pl.String:
            field_type = str
        elif dtype == pl.Datetime:
            field_type = datetime
        elif dtype == pl.Struct:
            field_type = dict[Any, Any]
        elif dtype == pl.List:
            if isinstance(cell[0], int):
                field_type = list[int]
            elif isinstance(cell[0], float):
                field_type = list[float]
            elif isinstance(cell[0], str):
                field_type = list[str]
            elif isinstance(cell[0], datetime):
                field_type = list[datetime]
            else:
                field_type = list[Any]

        if non_nulls.shape[0] < data.shape[0]:
            fields[column] = (field_type | None, ...)
        else:
            fields[column] = (field_type, ...)

    return await _serialize_data(model_name=table_name, data=data.to_dicts(), fields=fields)


async def serialize_table(table_name: str, data: pd.DataFrame | pl.DataFrame) -> list[BaseModel]:
    table_name = await camel_case(name=table_name)
    if isinstance(data, pd.DataFrame):
        data = pl.from_pandas(data=data)

    return await _serialize_table_polars(table_name=table_name, data=data)



# mypy: disable_error_code="type-arg"

from typing import Any

from polars import DataFrame
from pymongo import MongoClient


def find_in_collection(query: dict[str, Any], client: MongoClient, dbname: str, collection_name: str) -> DataFrame:
    db = client[dbname]

    collection = db[collection_name]
    documents = collection.find(query)

    return DataFrame(data=[document for document in documents])



# mypy: disable_error_code="type-arg"

from enum import Enum

from pymongo import AsyncMongoClient, MongoClient
from sqlalchemy.engine import Engine, create_engine
from sqlalchemy.ext.asyncio import AsyncEngine, create_async_engine


class SQLDatabaseDialect(str, Enum):
    """SQL database flavors."""

    MYSQL = "mysql"
    POSTGRESQL = "postgresql"
    SQLSERVER = "mssql"


class SQLDatabaseDriver(str, Enum):
    AIOMYSQL = "aiomysql"
    ASYNCPG = "asyncpg"
    PYMSSQL = "pymssql"


class NoSQLDatabaseDialect(str, Enum):
    MONGODB = "mongodb"
    SCYLLADB = "scylladb"
    CASSANDRA = "cassandra"


class NoSQLDatabaseDriver(Enum):
    MOTOR = "motor"


SQL_DRIVERS = {
    SQLDatabaseDialect.MYSQL.value: SQLDatabaseDriver.AIOMYSQL.value,
    SQLDatabaseDialect.POSTGRESQL.value: SQLDatabaseDriver.ASYNCPG.value,
    SQLDatabaseDialect.SQLSERVER.value: SQLDatabaseDriver.PYMSSQL.value,
}

NOSQL_DRIVERS = {NoSQLDatabaseDialect.MONGODB.value: NoSQLDatabaseDriver.MOTOR.value}


def create_connection_string(
    user: str,
    password: str,
    port: int,
    dialect: SQLDatabaseDialect,
    host: str,
    dbname: str,
) -> str:
    dbinfo = dialect.value
    if dbinfo in SQL_DRIVERS:
        dbinfo = f"{dbinfo}+{SQL_DRIVERS[dialect]}"

    connection_string = f"{dbinfo}://{user}:{password}@{host}:{port}/{dbname}"

    return connection_string


async def async_sql_engine(
    user: str,
    password: str,
    port: int,
    dialect: SQLDatabaseDialect,
    host: str,
    dbname: str,
) -> AsyncEngine:
    url = create_connection_string(
        user=user,
        password=password,
        port=port,
        dialect=dialect,
        host=host,
        dbname=dbname,
    )
    return create_async_engine(url=url)


def sync_sql_engine(
    user: str,
    password: str,
    port: int,
    dialect: SQLDatabaseDialect,
    host: str,
    dbname: str,
) -> Engine:
    url = create_connection_string(
        user=user,
        password=password,
        port=port,
        dialect=dialect,
        host=host,
        dbname=dbname,
    )

    return create_engine(url=url)


def nosql_client(user: str, password: str, dialect: NoSQLDatabaseDialect, host: str) -> MongoClient:
    driver = "mongodb"
    # if host != "localhost" and host != "127.0.0.1":
    driver += "+srv"
    connection_string = f"{driver}://{user}:{password}@{host}/retryWrites=true&w=majority"
    client: MongoClient = MongoClient(connection_string)

    return client


async def async_nosql_client(user: str, password: str, dialect: NoSQLDatabaseDialect, host: str) -> AsyncMongoClient:
    driver = "mongodb"
    # if host != "localhost" and host != "127.0.0.1":
    driver += "+srv"
    connection_string = f"{driver}://{user}:{password}@{host}/retryWrites=true&w=majority"
    client: AsyncMongoClient = AsyncMongoClient(connection_string)

    return client



# mypy: disable_error_code="type-arg"

import pytest
from bson.objectid import ObjectId
from polars import DataFrame
from pymongo import MongoClient

from univorm.reader.nosql import find_in_collection
from univorm.writer.nosql import insert_into_collection


@pytest.mark.asyncio
async def test_query_with_result(mongo_client: MongoClient) -> None:
    documents = [{"name": "test1"}, {"name": "test2"}]
    object_ids = insert_into_collection(documents=documents, client=mongo_client, dbname="test", collection_name="test")
    assert isinstance(object_ids, list)
    for object_id in object_ids:
        assert isinstance(object_id, ObjectId)

    df = find_in_collection(query={}, client=mongo_client, dbname="test", collection_name="test")
    assert isinstance(df, DataFrame)
    assert df.shape[0] > 0
    assert df.shape[1] > 1



from datetime import datetime

import pandas as pd
import polars as pl
import pytest
from pydantic import BaseModel

from univorm.parser import (
    camel_case,
    deserialize_pydantic_objects,
    flatten,
    serialize_table,
)


@pytest.mark.asyncio
async def test_camel_case() -> None:
    assert await camel_case(name="polars_table") == "PolarsTable"
    assert await camel_case(name="some_random.model") == "SomeRandomModel"


@pytest.mark.asyncio
async def test_serialize_pandas_table() -> None:
    now = datetime.now()
    data1 = {
        "n": "xin",
        "id": 200,
        "f": ["a", "b", "c"],
        "c": now,
        "b": 20.0,
        "d": {"a": 1},
        "e": [[1]],
    }
    data2 = {
        "n": "xin",
        "id": 200,
        "f": ["d", "e", "f"],
        "c": now,
        "b": None,
        "d": {"a": 1},
        "e": [[2]],
    }

    pydantic_objects = await serialize_table(table_name="some_table", data=pd.DataFrame(data=[data1, data2]))

    assert isinstance(pydantic_objects, list)
    for entry in pydantic_objects:
        assert entry.__class__.__name__ == "SomeTable"
        assert isinstance(entry, BaseModel)
        assert len(entry.model_dump().keys()) >= 5


@pytest.mark.asyncio
async def test_serialize_polars_table() -> None:
    now = datetime.now()
    data1 = {
        "n": "xin",
        "id": 200,
        "f": ["a", "b", "c"],
        "c": now,
        "b": 20.0,
        "d": {"a": 1},
        "e": [1],
        "d2": [now],
    }
    data2 = {
        "n": "xin",
        "id": 200,
        "f": ["d", "e", "f"],
        "c": now,
        "b": None,
        "d": {"a": 1},
        "g": [2.0],
    }

    df = pl.DataFrame([data1, data2])

    pydantic_objects = await serialize_table(table_name="polars_table", data=df)

    assert isinstance(pydantic_objects, list)
    for entry in pydantic_objects:
        assert entry.__class__.__name__ == "PolarsTable"
        assert isinstance(entry, BaseModel)
        assert len(entry.model_dump().keys()) >= 5


@pytest.mark.asyncio
async def test_flatten() -> None:
    data = [
        {
            "id": 1,
            "name": "Cole Volk",
            "fitness": {"height": 130, "weight": 60},
        },
        {"name": "Mark Reg", "fitness": {"height": 130, "weight": 60}},
        {
            "id": 2,
            "name": "Faye Raker",
            "fitness": {"height": 130, "weight": 60},
        },
    ]

    df_pandas = pd.DataFrame(data=data)
    df_polars = pl.DataFrame(data=data)

    flat_df = await flatten(data=df_pandas, depth=0)
    assert isinstance(flat_df, pl.DataFrame)
    assert flat_df.shape[1] == df_polars.shape[1]

    flat_df = await flatten(data=df_polars, depth=0)
    assert isinstance(flat_df, pl.DataFrame)
    assert flat_df.shape[1] == df_polars.shape[1]
    for col in flat_df.columns:
        assert col in df_polars.columns
    for col in df_polars.columns:
        assert col in flat_df.columns

    flat_df = await flatten(data=df_polars, depth=1)
    assert isinstance(flat_df, pl.DataFrame)
    assert flat_df.shape[1] == df_polars.shape[1] + 1
    assert "fitness" not in flat_df.columns
    assert "fitness.height" in flat_df.columns
    assert "fitness.weight" in flat_df.columns


@pytest.mark.asyncio
async def test_deserialize_models() -> None:
    class DummyModel(BaseModel):
        x: int
        y: float

    models = [DummyModel(x=1, y=2.0), DummyModel(x=10, y=-1.9)]
    df = await deserialize_pydantic_objects(models=models)

    assert isinstance(df, pl.DataFrame)
    assert df.shape[0] == 2
    assert df.shape[1] == 2



import pytest

from univorm.db import (
    NOSQL_DRIVERS,
    SQL_DRIVERS,
    NoSQLDatabaseDialect,
    NoSQLDatabaseDriver,
    SQLDatabaseDialect,
    SQLDatabaseDriver,
)


@pytest.mark.asyncio
async def test_dialects_drivers() -> None:
    for dialect in SQLDatabaseDialect:
        assert dialect.value in SQL_DRIVERS
        assert SQL_DRIVERS[dialect.value] in SQLDatabaseDriver

    assert NoSQLDatabaseDialect.MONGODB.value in NOSQL_DRIVERS
    assert NOSQL_DRIVERS[NoSQLDatabaseDialect.MONGODB.value] in NoSQLDatabaseDriver
